/* rfbproxy - a record/playback VNC proxy
 * Copyright (C) 2000  Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * This program is _very_ quick and dirty.  Please make it better.  It
 * works by recording everything the server says, and then playing it
 * back (really).  What would be better would be to actually take part
 * in the initialisation, and only start recording after that.
 *
 * I've found that hextile encoding gives the best results.  It speeds
 * it up quite a lot to make a hacked version of vncviewer that
 * just reads from the log file: that would cut out a write, a read,
 * and a context switch for each packet.  See the vncviewer patch in
 * this tarball..
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <sys/mman.h>
#include <fcntl.h>

#define BUFSIZE 65000

int record (const char *file)
{
	FILE *f = fopen (file, "w");
	int sockb = -1, sockv = -1, socks = -1;
	struct sockaddr_in sin;
	struct timeval last_tv;
	int first = 1;
	char *buf = malloc (BUFSIZE);
	int on = 1;

	if (!buf)
		goto out;

	sockb = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
	if (sockb == -1) {
		perror ("socket");
		goto out;
	}
	socks = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
	if (socks == -1) {
		perror ("socket");
		goto out;
	}

	setsockopt (sockb, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (int));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons (5910);
	if (bind (sockb, (struct sockaddr *)&sin, sizeof (struct sockaddr_in)))
	{
		perror ("bind");
		goto out;
	}

	printf ("Ready\n");
	listen (sockb, 1);
	sockv = accept (sockb, NULL, 0);
	close (sockb);
	sockb = -1;
	printf ("Accepted connection\nForwarding..");
	sin.sin_port = htons (5901);
	if (connect (socks, (struct sockaddr *)&sin,
		     sizeof (struct sockaddr_in))) {
		perror (" ");
		goto out;
	}

	printf (" connected\n");

	for (;;) {
		ssize_t bufs;
		struct timeval tv;
		struct timezone tz;
		fd_set rfds;
		FD_ZERO (&rfds);
		FD_SET (socks, &rfds);
		FD_SET (sockv, &rfds);
		if (select(FD_SETSIZE, &rfds, NULL, NULL, NULL) < 0) {
			perror ("select");
			goto out;
		}

		if (FD_ISSET (socks, &rfds)) {
			struct timeval diff;
			gettimeofday (&tv, &tz);
			if (first) {
				first = 0;
				last_tv = tv;
			}
			bufs = read (socks, buf, BUFSIZE);
			if (!bufs) break;
			write (sockv, buf, bufs);
			diff.tv_sec = tv.tv_sec - last_tv.tv_sec;
			diff.tv_usec = tv.tv_usec - last_tv.tv_usec;
			while (diff.tv_usec < 0) {
				diff.tv_sec--;
				diff.tv_usec += 1000000;
			}
			while (diff.tv_usec > 999999) {
				diff.tv_sec++;
				diff.tv_usec -= 1000000;
			}
			fwrite (&diff, sizeof (diff), 1, f);
			fwrite (&bufs, sizeof (bufs), 1, f);
			fwrite (buf, 1, bufs, f);
			last_tv = tv;
		}
		else {
			bufs = read (sockv, buf, BUFSIZE);
			if (!bufs) break;
			write (socks, buf, bufs);
		}
	}
 out:
	free (buf);
	if (sockb != -1)
		close (sockb);
	if (sockv != -1)
		close (sockv);
	if (socks != -1)
		close (socks);
	fclose (f);
	return 0;
}

int playback (const char *file)
{
	int fd = open (file, O_RDONLY);
	int sockb = -1, sockv = -1;
	struct sockaddr_in sin;
	struct stat st;
	char *map, *buf;
	char *skip;
	int on = 1;

	fstat (fd, &st);
	map = buf = mmap (NULL, st.st_size, PROT_READ | PROT_WRITE,
			  MAP_PRIVATE, fd, 0);
	close (fd);
	if (!map) return 1;

	skip = malloc (BUFSIZE);
	if (!skip)
		goto out;

	sockb = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
	if (sockb == -1) {
		perror ("socket");
		goto out;
	}

	setsockopt (sockb, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (int));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons (5910);
	if (bind (sockb, (struct sockaddr *)&sin, sizeof (struct sockaddr_in)))
	{
		perror ("bind");
		goto out;
	}

	printf ("Ready\n");
	listen (sockb, 1);
	sockv = accept (sockb, NULL, 0);
	close (sockb);
	sockb = -1;
	printf ("Accepted connection\nFaking..\n");

	for (;;) {
		ssize_t bufs;
		struct timeval *tv;
		fd_set rfds;
		FD_ZERO (&rfds);
		FD_SET (sockv, &rfds);

		if (map + st.st_size - buf < sizeof (tv) + sizeof (bufs))
			break;
		tv = (struct timeval *) buf; buf += sizeof (*tv);
		bufs = * (ssize_t *) buf; buf += sizeof (ssize_t);

		while (select(FD_SETSIZE, &rfds, NULL, NULL, tv) != 0) {
			if (!read (sockv, skip, BUFSIZE)) {
				/* Connection closed. */
				close (sockv);
				goto out;
			}
		}

		write (sockv, buf, bufs);
		buf += bufs;
	}
	close (sockv);
 out:
	free (skip);
	munmap (map, st.st_size);
	if (sockb != -1)
		close (sockb);
	if (sockv != -1)
		close (sockv);
	return 0;
}

void usage (const char *name)
{
	fprintf (stderr,
		 "usage: %s --record <log>\n"
		 "       %s --playback <log>\n",
		 name, name);
	exit (1);
}

int main (int argc, char *argv[])
{
	if (argc < 2) {
		usage (argv[0]);
	}

	/* Very simple arg parsing */
	if (!strcmp (argv[1], "--record")) {
		return record (argv[2]);
	}

	if (!strcmp (argv[1], "--playback")) {
		return playback (argv[2]);
	}

	usage (argv[0]);
	return 1;
}
